/* Pac-Man for Terminals */

#include <stdio.h>
#include <stdlib.h>
#include "newvt.h"
#include "pacman.h"
#include "pac-types.h"
#include "clock.h"


void init_ghosts(void)
{
    int g,j;

    for (g=0;g<PL.numghosts;g++)
    {
        PG.state    =Init;
        PG.inhome   =1;
        PG.loc      =PL.g_home[g];
        PG.goal     =PL.g_init[g];
        PG.dir      =Up;
        PG.mv.l_move=0;
        PG.mv.speed =(CLOCKS_PER_SEC*10)/PL.g_speed[g];
        for (j=0;j<MEMORY;PG.mem[j++]=point(-1,-1));
        PG.mp       =0;
    }
}

void reset_ghosts(void)
{
    int g,j;

    for (g=0;g<PL.numghosts;g++)
    {
        PG.state    =Init;
        PG.inhome   =1;
        PG.loc      =PL.g_home[g];
        PG.goal     =PL.g_init[g];
        PG.dir      =Up;
        for (j=1;j<5;j++) if (pass(PG.loc,j,Init)) 
        { 
            PG.dir  =j; 
            break; 
        }
        PG.mv.l_move=Clock();
        PG.mv.speed =(CLOCKS_PER_SEC*10)/PL.g_speed[g];
        for (j=0;j<MEMORY;PG.mem[j++]=point(-1,-1));
        PG.mp       =0;
    }
}

void init_pacman(void)
{
    for (up=0;up<NUMPLAYERS;up++)
    {
        init_level(0);
        PP.rand     =0;
        PP.god      =0;
        PP.power    =0;
        PP.fpower   =0;
        PP.score    =0;
        PP.lives    =LIVES;
        PP.loc      =PL.p_home;
        PP.dir      =Left;
        PP.ndir     =Left;
        PP.mv.l_move=Clock();
        PP.mv.speed =(CLOCKS_PER_SEC*10)/PL.p_speed;
        PP.mv.cspeed=(CLOCKS_PER_SEC*10)/PL.pc_speed;
        PP.time_left=1;
    }

    up=0;
}       

void reset_pacman(void)
{
    PP.rand     =0;
    PP.loc      =PL.p_home;
    PP.dir      =Left;
    PP.ndir     =Left;
    PP.power    =0;
    PP.fpower   =-1;
    PP.time_left=1;
}

int is_tunnel(TMazeLoc pt)
{
    if (!(pt.x<=3 || pt.x>=(PL.mazex-4)))  return 0;
    if (pass(pt,Up,Active) || pass(pt,Down,Active)) return 0;
    if (pass(point(2,pt.y),Left,Active) && 
        pass(point(PL.mazex-3,pt.y),Right,Active)) return 1;

    return 0;
}

void move_ghost(int g)
{
    /* Ghost AI */

    /* Are we in an intersection?
     *
     * Yes: Pick new dir that will further us towards our "goal"
     *      without backtracking.
     *
     * No:  Keep moving along corridor.
     */

    int i,j,k,dir;
    int rdir;
    TMazeLoc n;

    if (coinc(PG.loc,PL.g_home[g])) PG.inhome++; 

    for (i = g-1; i>0; i--)
    {
        if (P.ghost[i].inhome)
            PG.inhome = 0;
    }

    rdir=PG.dir==Up?Down:
        (PG.dir==Down?Up:
        (PG.dir==Left?Right:Left));

    n=PG.loc;

    i = dir = 0;

    if (PG.inhome > 4)
    for (j=1; j<=4; j++)
    {
        TMazeLoc pt;

        pt=pdir(PG.loc,j);

        if (PL.map[idx(pt)] == 'O' && PL.map[idx(pt)+1] == 'O')
        {
            dir |= 1<<j; 
            i++;
        }
    }

    if (PG.inhome > 4 && PG.state == Active)    
        PG.state = Dead;    /* fix race */

    if (!i)
    for (dir=0,j=1;j<=4;j++)
    {   
        if (pass(n,j,PG.inhome?Dead:PG.state))
        {
            dir |= 1<<j;
            i++;
        }
    }

        
    /*go(2,1); printf("i=%d dir=%x",i,dir); fflush(stdout); go(1,1);*/
     

    if (i==0)
    {
        fprintf(stderr,"move_ghost(): ghost painted into a corner?!\n");
        exit(1);
    }

    if (i==1)
    {
        for (j=0,i=1;i<=4;i++) if (dir&(1<<i)) {j=i; break;}
        PG.dir=dir=j;
    } else if (i==2)
    {
        dir&=~(1<<rdir);
        for (j=0,i=1;i<=4;i++) if (dir&(1<<i)) {j=i; break;}
        if (j==0)
        {
            fprintf(stderr,"ghost is confused.\n");
            exit(1);
        }
        PG.dir=dir=j;
    } else /* intersection */
    {
        int dirs[4];
        int a,b,c,d;
        TMazeLoc w;
        dir&=~(1<<rdir);

        w=pdiff(PG.goal,PG.loc);

        if (w.x==0)
        {
            a=Up; d=Down; 
            if (g&1) { b=Left; c=Right; } else { b=Right; c=Left; }

            if (w.y<0)
            {
                dirs[0]=a; dirs[1]=b; dirs[2]=c; dirs[3]=d;
            } else
            {
                dirs[0]=d; dirs[1]=c; dirs[2]=b; dirs[3]=a;
            }
        } else if (w.y==0)
        {
            a=Left; d=Right;
            if (g&2) { b=Up; c=Down; } else { b=Down; c=Up; }

            if (w.x<0)
            {
                dirs[0]=a; dirs[1]=b; dirs[2]=c; dirs[3]=d;
            } else {
                dirs[0]=d; dirs[1]=c; dirs[2]=b; dirs[3]=a;
            }
        } else
        {
            if (w.x<0) { a=Left; b=Right; w.x=-w.x; }
            else       { b=Left; a=Right; }

            if (w.y<0) { c=Up; d=Down; w.y=-w.y; }
            else       { d=Up; c=Down; }

            if (w.x>w.y) { dirs[0]=a; dirs[1]=c; dirs[2]=d; dirs[3]=b; }
            else         { dirs[0]=c; dirs[1]=a; dirs[2]=b; dirs[3]=d; }
        }

        if (PG.state==Chased)
        {
            for (i=3,j=0;i>=0;i--)
            {
                if (dirs[i]==rdir || (dir&(1<<dirs[i]))==0) continue;
                for (k=0;k<MEMORY;k++) 
                    if (coinc(PG.mem[k],pdir(n,dirs[i]))) break;
                if (k!=MEMORY) continue;
                if (pass(n,dirs[i],PG.inhome?Dead:PG.state)) 
                { 
                    int og = g;

                    j=dirs[i]; 
                    /*PG.mem[PG.mp]=pdir(n,j);*/
                    /*PG.mp=(PG.mp+1)%MEMORY;*/

                    for (g = 0; g <= (og + 1) && g < PL.numghosts; g++)
                        if (PG.state == Chased)
                        {
                            PG.mem[PG.mp]=pdir(n,j);
                            PG.mp=(PG.mp+1)%MEMORY;
                        }

                    g = og;

                    break;
                }
            }
            if (j==0)
            {
                if (rand()&1)
                {
                    int t;

                    t=dirs[0]; dirs[0]=dirs[1]; dirs[1]=t;
                    t=dirs[2]; dirs[2]=dirs[3]; dirs[3]=t;
                }
                if (rand()&2)
                {
                    int t;

                    t=dirs[0]; dirs[0]=dirs[2]; dirs[2]=t;
                    t=dirs[1]; dirs[1]=dirs[3]; dirs[3]=t;
                }
                for (i=3,j=0;i>=0;i--)
                {
                    if (dirs[i]==rdir || (dir&(1<<dirs[i]))==0) continue;
                    if (pass(n,dirs[i],PG.inhome?Dead:PG.state)) 
                        { j=dirs[i]; break;}
                }
            }
        } else
        {
            if (PG.state != Init) 
            {
                for (i=0,j=0;i<4;i++)
                {
                    if (dirs[i]==rdir || (dir&(1<<dirs[i]))==0) continue;

                    for (k=0;k<MEMORY;k++) 
                        if (coinc(PG.mem[k],pdir(n,dirs[i]))) break;

                    if (k!=MEMORY) continue;

                    if (pass(n,dirs[i],PG.inhome?Dead:PG.state)) 
                    {
                        int og = g;

                        j=dirs[i];
                        /*PG.mem[PG.mp]=pdir(n,j);*/
                        /*PG.mp=(PG.mp+1)%MEMORY;*/

                        for (g = 0; g <= (og + 1) && g < PL.numghosts; g++)
                            if (PG.state == P.ghost[g].state)
                            {
                                PG.mem[PG.mp]=pdir(n,j);
                                PG.mp=(PG.mp+1)%MEMORY;
                            }

                        g = og;
                        break;
                    }
                } 
                if (j==0 && rand()&1)
                {
                    int t;
        
                    t=dirs[0]; dirs[0]=dirs[1]; dirs[1]=t;
                    t=dirs[2]; dirs[2]=dirs[3]; dirs[3]=t;
                }
                if (j==0 && rand()&2)
                {
                    int t;
        
                    t=dirs[0]; dirs[0]=dirs[2]; dirs[2]=t;
                    t=dirs[1]; dirs[1]=dirs[3]; dirs[3]=t;
                }
            } else j=0;
            if (j==0) for (i=0,j=0;i<4;i++)
            {
                if (dirs[i]==rdir || (dir&(1<<dirs[i]))==0) continue;
                if (pass(n,dirs[i],PG.inhome?Dead:PG.state)) 
                    { j=dirs[i]; break;}
            }
        }
        PG.dir=dir=j;
    }

    n=pdir(n,dir);
    if (n.x>PL.mazex && dir==Right) { n.x=-2; }
    if (n.x<0        && dir==Left)  { n.x=PL.mazex; }

    if (!pass(n,0,PG.inhome?Dead:PG.state))
    {
        go(1,1); printf("move_ghost(): picked a non-passible direction!\n");
        fflush(stdout);
    }
    PG.loc=n;

    if (PG.state==Active || PG.state==Chased) 
    { 
        PG.goal=PP.loc;
    } else if (PG.state==Dead)
    {
        PG.goal=PL.g_home[g];
    }

    if (PL.map[idx(PG.loc)]=='O')
    {
        if (PG.state!=Dead) PG.inhome=0;
        if (PG.state==Dead) PG.inhome=1;
    }

    if (coinc(PG.loc,PG.goal) || 
        (PG.state==Dead && PL.map[idx(PG.loc)]=='O'))
    {
        switch (PG.state)
        {
            case Dead:  PG.state =Init;
                        PG.goal  =PL.g_init[g];
                        PG.inhome=1;
                        break;

            case Init:  PG.state=Active;
                        PG.goal =PP.loc;
                        break;

            case Chased:
            case Active: break;

            default:    fprintf(stderr,"move_ghost():ghost in unknown state!\n");
                        break;
        }
    }

#if 0
    go(height-g-1,1); fflush(stdout);
    fprintf(stderr,"Ghost %2d:  %10s   @[%2d,%2d]  Goal[%2d,%2d]  ih=%2d\n",
            g,
            PG.state == Dead   ? "Dead" :
            PG.state == Init   ? "Init" :
            PG.state == Chased ? "Chased" :
            PG.state == Active ? "Active" : "Unknown",
            PG.loc.x, PG.loc.y, PG.goal.x, PG.goal.y, PG.inhome);
#endif

}

void move_pacman(void)
{
    int rdir,i,j,dir;
    TMazeLoc n;

    if (PP.rand)
    {   
        rdir=PP.dir==Up?Down:
            (PP.dir==Down?Up:
            (PP.dir==Left?Right:Left));
    
        n=PP.loc;

        for (dir=0,i=0,j=1;j<=4;j++) {  if (pass(n,j,Active)) { dir|=(1<<j); i++; } }
    
        if (i==1)
        {
            PP.ndir=rdir;
        } else if (i==2)
        {
            dir&=~(1<<rdir);
            for (j=0,i=1;i<=4;i++) if (dir&(1<<i)) {j=i; break;}
            PP.ndir=j;
        } else /* intersection */
        {
            dir&=~(1<<rdir);
            do
            {
                PP.ndir=1+(rand()&3);
            } while (!(dir&(1<<PP.ndir)));
        }

    }

    if (pass(PP.loc,PP.ndir,Active))
    {
        PP.dir=PP.ndir;
    } 
    

    if (pass(PP.loc,PP.dir,Active))
    {
        PP.loc=pdir(PP.loc,PP.dir);
    }

    if (PP.loc.x<0         && PP.dir==Left)  { PP.loc.x+=PL.mazex; }
    if (PP.loc.x>=PL.mazex && PP.dir==Right) { PP.loc.x-=PL.mazex; }

    return;
}


void show_pacman(void)
{
    int x,y;

    x=PP.loc.x-PL.sx;
    y=PP.loc.y-PL.sy;
    if (x<-1 || x>=PL.dw || y<0 || y>=PL.dh) return;
    if (PP.dir==Up || PP.dir==Right)
    {
        if (x>=0)      putpixel(x+PL.lc,y+PL.tr,'(');
        if (x+1<PL.dw) putpixel(x+1+PL.lc,y+PL.tr,'<');
    } else
    {
        if (x>=0)      putpixel(x+PL.lc,y+PL.tr,'>');
        if (x+1<PL.dw) putpixel(x+1+PL.lc,y+PL.tr,')');
    }   
}

void show_ghost(int g)
{
    int x,y,c1,c2;

    x=PG.loc.x-PL.sx; 
    y=PG.loc.y-PL.sy;
    if (x<1 || x>=PL.dw || y<0 || y>=PL.dh) return;

    switch (PG.state)
    {
        case Init: 
        case Active:
            c1=c2='M';
            break;
        case Chased:
            if (PP.power*5<PP.fpower && PP.power&4)
            {
                c1=c2='w';
            } else
            {
                c1=c2='W';
            }
            break;
        case Dead:
            c1=c2='o';
            break;
        default:
            fprintf(stderr,"show_ghost: ghost in unknown state!\n");
            exit(1);
    }

    if (x>=0)      putpixel(x+PL.lc,y+PL.tr,c1);
    if (x+1<PL.dw) putpixel(x+PL.lc+1,y+PL.tr,c2);
}

void hide_all(void)  /* pacman # n*/
{
    int g;
    int x,y;
    int xx,yy;

    
    for (g=0;g<PL.numghosts;g++)
    {
        x=(xx=PG.loc.x)-PL.sx; y=(yy=PG.loc.y)-PL.sy;

        if (x<0 || x>=PL.dw || y<0 || y>=PL.dh) continue;
        putpixel(x+PL.lc,y+PL.tr,PL.map[idx(point(xx,yy))]);
        if (x+1<PL.dw) putpixel(x+PL.lc+1,y+PL.tr,PL.map[idx(point(xx+1,yy))]);
    }

    x=(xx=PP.loc.x)-PL.sx; y=(yy=PP.loc.y)-PL.sy;
    if (x<0 || x>=PL.dw || y<0 || y>=PL.dh) return;

    putpixel(x+PL.lc,y+PL.tr,PL.map[idx(point(xx,yy))]);
    if (x+1<PL.dw) putpixel(x+PL.lc+1,y+PL.tr,PL.map[idx(point(xx+1,yy))]);
}


void kill_pacman(void)
{
    /* (< (= {= =- __ */
    /* >) =) =} -= __ */

    char *die1r="(({=_";
    char *die2r="<==-_";
    char *die1l=">==-_";
    char *die2l="))}-_";
    char *d1,*d2;
    int i,x,y;

    if (PP.dir==Left || PP.dir==Down) 
            { d1=die1l; d2=die2l; }
    else    { d1=die1r; d2=die2r; }


    x=PP.loc.x-PL.sx;
    y=PP.loc.y-PL.sy;
    if (x<-1 || x>=PL.dw || y<0 || y>=PL.dh) return;
    for (i=0; i<5; i++)
    {
        if (x>=0)      putpixel(x+PL.lc,y+PL.tr,d1[i]);
        if (x+1<PL.dw) putpixel(x+1+PL.lc,y+PL.tr,d2[i]);
        supdate();
        usleep(200000);
    }   

}
    

/* ======================================================================== */
/*  Please see the file "COPYING" for the complete GNU GPL License.         */
/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       */
/*  General Public License for more details.                                */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License       */
/*  along with this program; if not, write to the Free Software             */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               */
/* ======================================================================== */
/*                 Copyright (c) 1995-2000, Joseph Zbiciak                  */
/* ======================================================================== */
