/* Pac-Man for Terminals */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#ifndef BYTE_ORDER
#include <endian.h>
#include <bytesex.h>     /* if these don't work, #define BYTE_ORDER as nec. */
#endif
#include "newvt.h"
#include "clock.h"
#include "pac-types.h"
#include "pacman.h"
#include "keybd.h"
#include "actors.h"

#if !defined(BYTE_ORDER) && defined(__BYTE_ORDER)
#define BYTE_ORDER __BYTE_ORDER
#elseif !defined(BYTE_ORDER)
#error Please define BYTE_ORDER in pacman.h
#endif

static unsigned char * tga=NULL; 
static int * b_tga=NULL, tga_x = -1, tga_y = -1;

int prand(unsigned int i)
{
    int j;

    i^=0x2a3a4a5a;
    i=i*i+i;

    for (j=37;--j;) { i=(((i<<27)^(i<<29))&0x80000000)|((i>>1)&0x7fffffff); }

    return i;
}

int tga_to_txt(char * tga_file,int nx,int ny, unsigned char * txt,int bias)
{
    char *lookup=" .,:iI$@#";
    FILE * infile;
    unsigned char *c,*o;
    static short ox,oy;
    int x,y,xx,yy,k;
    int r,g,b,idx;


    if (tga==NULL && b_tga==NULL)
    {
        infile=fopen(tga_file,"r");

        if (!infile)
        {
            fprintf(stderr,"tga_to_txt:  Unable to open `%s'\n",tga_file);
            perror("fopen()");
            exit(1);
        }
    
        if (fseek(infile,12,0))
        {
            fprintf(stderr,"tga_to_txt:  Unable to seek into TGA header\n");
            perror("fseek()");
            exit(1);
        }
        if (fread(&ox,2,1,infile)!=1)
        {
            fprintf(stderr,"tga_to_txt:  Unable to read from TGA header (x)\n");
            perror("fread()");
            exit(1);
        }
        if (fread(&oy,2,1,infile)!=1)
        {
            fprintf(stderr,"tga_to_txt:  Unable to read from TGA header (y)\n");
            perror("fread()");
            exit(1);
        }
        if (BYTE_ORDER%10==1)
        {
            ox=((ox&0xFF)<<8)|((ox>>8)&0xFF);
            oy=((oy&0xFF)<<8)|((oy>>8)&0xFF);
        }
        if (fseek(infile,18,0))
        {
            fprintf(stderr,"tga_to_txt:  Unable to seek into TGA body\n");
            perror("fseek()");
            exit(1);
        }
        tga=(unsigned char *)malloc(sizeof(unsigned char)*ox*oy*3);
        if (!tga)
        {
            fprintf(stderr,"tga_to_txt:  Unable to allocate buffer for TGA file\n");
            perror("malloc()");
            exit(1);
        }
        if (fread(tga,ox*oy*3,1,infile)!=1)
        {
            fprintf(stderr,"tga_to_txt:  Unable to read from TGA body\n");
            perror("fread()");
            exit(1);
        }
        fclose(infile);
    }

    o=txt;
    if (b_tga && (tga_x != nx || tga_y != ny))
    {
        free(b_tga);
        b_tga = NULL;
    }
    if (b_tga==NULL)
    {
        tga_x = nx;
        tga_y = ny;
        b_tga=(int *)malloc(sizeof(int)*nx*ny);
        if (!b_tga)
        {
            fprintf(stderr,"Unable allocate temp TGA processing buffer\n");
            perror("malloc()");
            exit(1);
        }
        for (y=0;y<ny;y++) 
        for (x=0;x<nx;x++)
        {
            k=0;
            r=g=b=0;
            for(yy=(y*oy)/ny;yy<=((y+1)*oy)/ny;yy++)
            for(xx=(x*ox)/nx;xx<=((x+1)*ox)/nx;xx++) 
            {
                r+=tga[3*xx+0+(yy*3*ox)];
                g+=tga[3*xx+1+(yy*3*ox)];
                b+=tga[3*xx+2+(yy*3*ox)];
                k++;
            }
            b_tga[x+y*nx]=(11.4*b+58.7*g+29.9*r)/(100*k)+((prand(x+y*nx)&31)-16);   
            idx=(b_tga[x+y*nx]+bias)/(32);
            if (idx<0) idx=0; else if (idx>7) idx=7;
    
            o[x+y*nx]=lookup[idx];
        }
    } else 
    {
        for (x=0;x<nx*ny;x++) 
        {
            idx=(b_tga[x]+bias)/(32);
            o[x]=lookup[idx<0?0:idx>7?7:idx]; 
        }
    }

    return 0;
}

static int tx_w[257], tx_h[257];
static unsigned char *txt=NULL, *txts[257];
static int in_out=0;



void fade_in(int length)
{
    int i,j;

    if (txt==NULL) for (i=0;i<257;i++) txts[i]=NULL;

    if (in_out==1) return;
    for (j=-256;j<=0;j+=(256/length))
    {
        if (j>0) j=0;  if (j<-256) j=-256;
        if (txts[-j] != NULL && (tx_w[-j] != width || tx_h[-j] != height))
        {
            free(txts[-j]);
            txts[-j] = NULL;
        }
        if (txts[-j]==NULL) 
        {
            txts[-j]=(unsigned char *)malloc(width*height);
            tga_to_txt("pacman.tga",width,height,txts[-j],j);
            tx_w[-j] = width;
            tx_h[-j] = height;
        }
        txt=txts[-j];
        for (i=0;i<width*height;i++) screen[i]=txt[i];
        supdate();
        usleep(20000);
    }
    if (txts[0] && (tx_w[0] != width || tx_h[0] != height))
    {
        free(txts[0]);
        txts[0] = NULL;
    }
    if (txts[0]==NULL) 
    {
        txts[0]=(unsigned char *)malloc(width*height);
        tga_to_txt("pacman.tga",width,height,txts[0],0);
        tx_w[0] = width;
        tx_h[0] = height;
    }
    txt=txts[0];
    for (i=0;i<width*height;i++) screen[i]=txt[i];
    supdate();
    sync_via_rcp();
    in_out=1;
}

void fade_out(int length)
{
    int i,j;

    if (txt==NULL) for (i=0;i<257;i++) txts[i]=NULL;

    if (in_out==0) return;
    for (j=0;j>=-256;j-=(256/length))
    {
        if (j>0) j=0;  if (j<-256) j=-256;
        if (txts[-j] && (tx_w[-j] != width || tx_h[-j] != height))
        {
            free(txts[-j]);
            txts[-j] = NULL;
        }
        if (txts[-j]==NULL) 
        {
            txts[-j]=(unsigned char *)malloc(width*height);
            tga_to_txt("pacman.tga",width,height,txts[-j],j);
            tx_w[-j] = width;
            tx_h[-j] = height;
        }
        txt=txts[-j];
        for (i=0;i<width*height;i++) screen[i]=txt[i];
        supdate();
        usleep(20000);
    }
    if (txts[256] && (tx_w[256] != width || tx_h[256] != height))
    {
        free(txts[256]);
        txts[256] = NULL;
    }
    if (txts[256]==NULL) 
    {
        txts[256]=(unsigned char *)malloc(width*height);
        tga_to_txt("pacman.tga",width,height,txts[256],-256);
        tx_w[256] = width;
        tx_h[256] = height;
    }
    txt=txts[256];
    for (i=0;i<width*height;i++) screen[i]=txt[i];
    supdate();
    sync_via_rcp();
    in_out=0;
}

int title_screen(void)
{
    static int first_time = 1;
    int i,j,k,l,r,m,kb;
    char *msg2=" Press enter to start ",key;
    char *msg1="Joe Zbiciak,  Jun 2000";  /* Originally written Nov 1995 */
    int s;

    can_resize = 0;

    if (first_time)
    {
        for (i=0;i<257;i++)
            txts[i] = NULL;
        first_time = 0;
    }

    in_out=0;
    flushkbd();

    if (!txt)
    {
        fade_in(32);
        if (!txt)
        {
            fprintf(stderr,"title_screen:  Unable to alloc title screen buffer\n");
            perror("malloc()");
            exit(1);
        }
    } else { fade_in(32); }

    r=(height-8)/2;
    l=strlen(logo[0]);
    k=(width-l)/2;

    for(j=0;j<l;j++) { screen[(r+4)*width+j+k]=' '; }
    for(i=0;i<4;i++) for(j=0;j<l;j++)
        screen[(r+i)*width+j+k]=logo[i][j];

    usleep(500000);
    supdate();

    strncpy((char*)screen+(r+5)*width+(width-13)/2,"             ",13);
    strncpy((char*)screen+(r+6)*width+(width-13)/2," For VT-100! ",13);
    strncpy((char*)screen+(r+7)*width+(width-13)/2,"             ",13);
    usleep(400000);
    supdate();

    usleep(250000);

    s=(height-2)*width+(width-22)/2;
    init_pacman();
    reset_ghosts();
    reset_pacman();
    P.lv=0;
    init_level(0);
    kb=0;
    while (kb==0 && (readkbd(&key),kb=key=='\r'||key=='\n')==0) 
    {
        s=(height-2)*width+(width-22)/2;
        r=(height-8)/2;
        k=(width-l)/2;

        strncpy((char*)screen+(r+5)*width+(width-13)/2,"             ",13);
        strncpy((char*)screen+(r+6)*width+(width-13)/2," For VT-100! ",13);
        strncpy((char*)screen+(r+7)*width+(width-13)/2,"             ",13);
        strncpy((char*)screen+s-2*width,msg1,22);
        for(m=0;m<5;m++)
        {
            if ((kb=(readkbd(&key),key=='\r'||key=='\n'))!=0) break;
            for(i=0;i<5;i++) for(j=0;j<l;j++)
                screen[(r+i)*width+j+k]=txt[(r+i)*width+j+k];
            strncpy((char*)screen+s,(char*)txt+s,22);
            supdate();
            sync_via_rcp();
            usleep(500000);

            if ((kb=(readkbd(&key),key=='\r'||key=='\n'))!=0) break;
            for(j=0;j<l;j++) { screen[(r+4)*width+j+k]=' '; }
            for(i=0;i<4;i++) for(j=0;j<l;j++)
                screen[(r+i)*width+j+k]=logo[i][j];
            strncpy((char*)screen+s,msg2,22);
            supdate();
            sync_via_rcp();
            usleep(500000);
        }
        if (kb!=0 || (kb=(readkbd(&key),key=='\r'||key=='\n'))!=0) break;
        fade_out(8);
        reset_ghosts();
        reset_pacman();
        PP.rand=1;
        PP.time_left=500;
        PP.lives=0;
        flushkbd();

        if ((m=game_loop()) < 0) break;

        if (m==1) 
        {
            kill_pacman();
        }
        if (!PL.dots_left) 
        {
            close_level(P.lv);
            P.lv++;
            if (P.lv>=NUMLEVELS) P.lv=0;
            init_level(P.lv);
        }
        fade_in(8);
        flushkbd();
    }

    init_pacman();

    fade_out(32);

    for (i=0;i<257;i++) if (txts[i]) { free(txts[i]); txts[i]=NULL; }
    /*for(i=0;i<width*height;i++) screen[i]=' ';*/

    can_resize = 1;

    return m;
}

/* ======================================================================== */
/*  Please see the file "COPYING" for the complete GNU GPL License.         */
/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       */
/*  General Public License for more details.                                */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License       */
/*  along with this program; if not, write to the Free Software             */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               */
/* ======================================================================== */
/*                 Copyright (c) 1995-2000, Joseph Zbiciak                  */
/* ======================================================================== */
