/* Pac-Man for Terminals */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <termios.h>
#include "clock.h"
#include "pac-types.h"


int keybd, old_kbd_flags;
struct termios old_kbd_termio, kbd_termio;

void restorekbd(void)
{
    fcntl(keybd,F_SETFL,old_kbd_flags);
    tcsetattr(keybd,TCSANOW,&old_kbd_termio);
}

void initkbd(void)
{
    keybd=dup(0);

    old_kbd_flags=fcntl(keybd,F_GETFL);
    /*fcntl(keybd,F_SETFL,O_NDELAY);*/

    tcgetattr(keybd,&old_kbd_termio);

    kbd_termio=old_kbd_termio;
    kbd_termio.c_iflag   &=~(IXON|IXOFF);
    kbd_termio.c_lflag   &=~(ECHO|ECHOE|ECHOK|ECHONL|ICANON);
    kbd_termio.c_cc[VMIN] =0;
    kbd_termio.c_cc[VTIME]=0;
    tcsetattr(keybd,TCSANOW,&kbd_termio);

    atexit(restorekbd);
    return;
}

void die(int s)
{
    restorekbd();
    printf("\nExiting on signal %d   \n",s);
    return;
}

void flushkbd(void)
{
    tcflush(keybd,TCIFLUSH);
}

int readkbd(char *c)
{
    return read(keybd,c,1);
}

void slowkbd(void)
{
    kbd_termio.c_cc[VMIN] =1;
    kbd_termio.c_cc[VTIME]=10;
    tcsetattr(keybd,TCSANOW,&kbd_termio);
    return;
}

void fastkbd(void)
{
    kbd_termio.c_cc[VMIN] =0;
    kbd_termio.c_cc[VTIME]=0;
    tcsetattr(keybd,TCSANOW,&kbd_termio);
    return;
}


enum { NO_SYNC=0, BG_SYNC, HARD_SYNC, SOFT_SYNC, KITCHEN_SYNC };

#define MAXQUEUE 15
int kq=0,keyqueue[MAXQUEUE];
const char *keys="\033[R";
int sync_type=BG_SYNC, rcv_sync=0;

void send_rcp(void)
{
    fprintf(stderr,"\033[6n"); fflush(stderr);
}

void
sync_via_rcp
(void)
{
    int adj,d;
    char c;

    slowkbd();
    adj=Clock_unad();

    send_rcp();

    while (readkbd(&c)==0) ;

    if (kq < MAXQUEUE && (c>='A' || c==' ') 
        && c!=EOF && !strchr(keys,c)) keyqueue[kq++]=c;

    while ((d=readkbd(&c)),1 && c!='R' && c!='\n' && c!='\r') 
    {
        if (d>0 && kq < MAXQUEUE && (c>='A' || c==' ') 
            && c!=EOF && !strchr(keys,c)) keyqueue[kq++]=c;
    }

    clock_ad+=Clock_unad()-adj;
    fastkbd();
}

int parse_key(void)
{
    int d,q=0;
    char c;
    static int wf=0,sync_tmp;



    if (PP.rand && rand()&255) { PP.ndir=1+(rand()&3); }
    while (kq>0 || readkbd(&c)==1)
    {
        if (kq) { c=keyqueue[q++]; }
        if (q==kq) kq=0;

        if (PP.rand && !strchr("0123456789\033[R;Hs~",c)) 
            return c == 'Q' || c == 'q' ? -1 : -2; /* quit demo */


        if (c==('W') && !wf) 
        { 
            wf=1; 
            sync_tmp=sync_type; 
            sync_type=KITCHEN_SYNC; 
        } else if (wf && c=='R') 
        {
            wf=2;
            continue;
        } else if (wf)         
        { 
            int ret=0;

            if (wf==2 && c>='0' && c<='9') ret=-7-(c-'0'); 
            if (!strchr("\033[R;H~",c))
            {
                wf=0; 
                sync_type=sync_tmp; 
            }
            if (ret) return ret;
        }

        switch (c) 
        {
#ifdef APPLE_KEYS
            case 'i': case 'I':     PP.ndir=Up;                     break;
            case 'm': case 'M':     PP.ndir=Down;                   break;
            case 'j': case 'J':     PP.ndir=Left;                   break;
            case 'k': case 'K':     PP.ndir=Right;                  break;
#else               
            case 'k': case 'K': case 'A':   PP.ndir=Up;             break;
            case 'j': case 'J': case 'B':   PP.ndir=Down;           break;
            case 'h':           case 'D':   PP.ndir=Left;           break;
            case 'l': case 'L': case 'C':   PP.ndir=Right;          break;
#endif
            case 'g':               PP.god=!PP.god;                 break;
            case 'R':               rcv_sync=1;                     break;
            case 's': case 'S':     sync_type=(sync_type+1)%4;      break;
            case 'q': case 'Q':     return -1; /*quit*/             break;
            case '~':               showstats=!showstats;           break;
            case '*':
            {
                int g;
                for (g = 0; g < PL.numghosts; g++)
                    if (PG.state == Active || PG.state == Init) PG.state=Dead;
            } break;
        }
    }


    return 0;

}
/* ======================================================================== */
/*  Please see the file "COPYING" for the complete GNU GPL License.         */
/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       */
/*  General Public License for more details.                                */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License       */
/*  along with this program; if not, write to the Free Software             */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               */
/* ======================================================================== */
/*                 Copyright (c) 1995-2000, Joseph Zbiciak                  */
/* ======================================================================== */
