/* Pac-Man for Terminals */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "newvt.h"
#include "pac-types.h"
#include "pacman.h"
#include "actors.h"
#include "clock.h"

void init_level(int l)
{
    memcpy((void *)&(P.lev),(void *)&(level[l]),sizeof(TLevel));
}

void get_line(FILE * f, char * buf, int size)
{
    while (!feof(f))
    {
        if (!fgets(buf,size,f))
        {
            fprintf(stderr,"load_levels():get_line(): Unexpected EOF!\n");
            perror("fgets()");
        }
        if (buf[0]!=';') break;
    }
}


void load_levels(void)
{
    int i,j,k;
    unsigned char *c;
    FILE * f;
    char buf[100];

    for (i=0;i<NUMLEVELS;i++)
    {
        sprintf(buf,"level.%d",i);
        f=fopen(buf,"r");
        if (!f)
        {
            fprintf(stderr,"Error opening '%s'.\n",buf);
            perror("load_levels():fopen()");
            exit(1);
        }       

/*fprintf(stderr,"Loading '%s'\n",buf); fflush(stderr);*/

        get_line(f,buf,100);
        sscanf(buf,"%d %d",&(level[i].mazex),&(level[i].mazey));
/*fprintf(stderr,"Maze dims: %d %d\n",level[i].mazex,level[i].mazey); fflush(stderr);*/

        get_line(f,buf,100);
        sscanf(buf,"%d",&(level[i].numghosts));
/*fprintf(stderr,"Number of ghosts: %d\n",level[i].numghosts); fflush(stderr);*/

        for (j=0;j<level[i].numghosts;j++)
        {
            get_line(f,buf,100);
            sscanf(buf,"%d %d %d %d %d",    &(level[i].g_home[j].x),
                                            &(level[i].g_home[j].y),
                                            &(level[i].g_init[j].x),
                                            &(level[i].g_init[j].y),
                                            &(level[i].g_speed[j]));
/*fprintf(stderr,"Ghost %d: g_home(%d %d) g_init(%d %d) g_speed(%d)\n",j,
                level[i].g_home[j].x,level[i].g_home[j].y,
                level[i].g_init[j].x,level[i].g_init[j].y,
                level[i].g_speed[j]);  fflush(stderr);*/
        }

        get_line(f,buf,100);
        sscanf(buf,"%d %d %d %d",   &(level[i].p_home.x),
                                    &(level[i].p_home.y),
                                    &(level[i].p_speed), 
                                    &(level[i].pc_speed));
/*fprintf(stderr,"Pacman: p_home(%d %d) p_speed(%d)\n",
                level[i].p_home.x, level[i].p_home.y, level[i].p_speed); fflush(stderr);*/

        get_line(f,buf,100);
        buf[99]=0;
        sscanf(buf,"%d",&(level[i].power_up));
        c=level[i].map;
        for (j=0;j<level[i].mazey;j++)
        {
            if (feof(f)) 
            {
                fprintf(stderr,"load_levels():Unexpected EOF on map file.\n");
                exit(1);
            }
            get_line(f,buf,100);
            for (k=strlen(buf)-1;k<=level[i].mazex+1 && k<100;buf[k++]=' ');
            
            memcpy(c,buf,level[i].mazex);
/*fprintf(stderr,"line %2d: %s",j,buf); fflush(stderr);*/
            c+=level[i].mazex;
        }
        fclose(f);
        /* scan level for dots and powerups */

        level[i].tot_dots=level[i].tot_power=0;
        for (j=0;j<level[i].mazex*level[i].mazey;j++)
        {
            switch(level[i].map[j])
            {
            case DOT:   
                level[i].tot_dots++;    
                break;
            case POWER: 
                level[i].power_ups[level[i].tot_power++]
                        =point(j%level[i].mazex,j/level[i].mazex);
                break;
            }
        }
        level[i].power_left=level[i].tot_power;
        level[i].dots_left =level[i].tot_dots;
    }
}               

void open_level(int lev)
{
    int i,g,j,k;
    char *msg=" R E A D Y ! ";

    can_resize = 0;

    for (i=0;i<6;i++)
    {
        PL.sx=PL.sy=-1;
        show_map(0);
        if (!(i&1))
        {
            show_pacman();
            for (g=0;g<PL.numghosts;g++) show_ghost(g);
        } else
        {
            hide_all();
            k=((TM+height-BM)/2)*width + (LM+width-RM-13)/2;
            for (j=0;j<13;j++)
                screen[k++]=msg[j];
        }
        supdate();
        usleep(500000);
    }

    PL.sx=PL.sy=-1;
    show_map();
}

void game_over(int lev)
{
    int i,g,j,k;
    char *msg=" G A M E  O V E R ";
    char buf[19];

    can_resize = 0;
    hide_all();

    k=((TM+height-BM)/2)*width + (LM+width-RM-19)/2;
    for (i=0;i<19;i++) buf[i]=screen[k+i];

    for (j=0;j<19;j++) screen[k+j]=msg[j];
    supdate();
    sleep(1);
    for (i=0;i<100;i++)
    {
        usleep(125000);
        hide_all();
        for (j=0;j<19;j++) screen[k+j]=buf[j];
        for (g=0;g<PL.numghosts;g++) { move_ghost(g); show_ghost(g); }
        if (((2*Clock())/(CLOCKS_PER_SEC))&1)
                for (j=0;j<19;j++) screen[k+j]=msg[j];
        supdate();
    }

    for (i=0;i<width*height;i++) screen[i]=' ';

}

void close_level(int lev)
{
    int i,j,k;
    can_resize = 0;

    for (i=0;i<10;i++)
    {
        if (i&1) { PL.sx=PL.sy=-1; show_map(); }
        else 
        {
            for (j=TM;j<height-BM;j++) for (k=LM;k<width-RM;k++)
            {
                screen[j*width+k]=' ';
            }
        }

        supdate();
        usleep(500000);
    }
    for (j=TM;j<height-BM;j++) for (k=LM;k<width-RM;k++)
    {
        screen[j*width+k]=' ';
    }
}

/* ======================================================================== */
/*  Please see the file "COPYING" for the complete GNU GPL License.         */
/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       */
/*  General Public License for more details.                                */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License       */
/*  along with this program; if not, write to the Free Software             */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               */
/* ======================================================================== */
/*                 Copyright (c) 1995-2000, Joseph Zbiciak                  */
/* ======================================================================== */
