/* newvt */
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <fcntl.h>
#ifndef NORESIZE
#include <errno.h>
#include <signal.h>
#include <termio.h>
#endif

#ifdef NORESIZE
#undef VERSION
#undef AUTHOR
#undef TITLE
#define VERSION "v1.1-NR, Nov 1994"
#else
#define VERSION "v1.1-R, Nov 1994"
#endif
#define AUTHOR  "Joseph Zbiciak"
#define TITLE   "VT-100 Graphics Library"
#define S_DIRECT 0
#define S_DUMB   1
#define S_SMART  2
#define S_DELTA  3
#define S_SMARTB 4
#define S_DELTAB 5

unsigned char * screen;  /* screen display array */
unsigned char * oscreen; /* previous display page.  (used by update for efficient updates) */
int width, height, whsize; /* display dimensions */
int showstats=0; /* whether to show status info */
int frame=0, bytes=0; 
volatile int need_resize = 0, can_resize = 1;
int count;
int cx, cy;
#define XPARENT 1
#define abs(a) (((a)>=0)?(a):-(a))
#define sq(a) ((a)*(a))

#define go(x,y) bytes+=printf("\033[%d;%dH",cx=(x),cy=(y)) 

inline int sgo(char * result,int sx,int sy,int x, int y)
{
    if (y==1) {
        if (sx==x) {
            /*
            sprintf(result,"\r");
            return strlen(result);
            */
            *result='\r';
            *(result+1)=0;
            return 1;
        } else if ((sx+1)==x) {
            /*
            sprintf(result,"\n");
            return strlen(result);
            */
            *result='\n';
            *(result+1)=0;
            return 1;
        }
    }
    if (sx==x) {
        if (sy==y) {
            *result='\0';
            return 0;
        } else if (sy>y) {
            if (sy-y>1) 
                sprintf(result,"\033[%dD",sy-y); 
            else {
                result[0]='\033';
                result[1]='[';
                result[2]='D';
                return 3;
            }
        } else if (sy<y) {
            if (y-sy>1) {
                sprintf(result,"\033[%dC",y-sy); 
            } else {
                result[0]='\033';
                result[1]='[';
                result[2]='C';
                return 3;
            }
        }
        return strlen(result);
    } else if (sy==y) {
        if (sx==x) {
            *result='\0';
            return 0;
        } else if (sx>x) {
            if (sx-x>1) {
                sprintf(result,"\033[%dA",sx-x); 
                return strlen(result);
            } else {
                result[0]='\033';
                result[1]='[';
                result[2]='A';
                return 3;
            }
        } else if (sx<x) {
            if (x-sx>1) {
                sprintf(result,"\033[%dB",x-sx); 
                return strlen(result);
            } else {
                result[0]='\033';
                result[1]='[';
                result[2]='B';
                return 3;
            }
        }
        return strlen(result);
    }  
    if (x==1 && y==1) {
        result[0]='\033';
        result[1]='[';
        result[2]='H';
        return 3;
    } else if (x==1) {
        sprintf(result,"\033[;%dH",y);
    } else if (y==1) {
        sprintf(result,"\033[%d;H",x);
    } else {
        sprintf(result,"\033[%d;%dH",x,y);
    }
    return strlen(result);
}
#ifndef NORESIZE
void do_resize(void);
void get_dimz(int *w, int *h);
#endif

void supdate(void)
{
    int h,i,j,k,l,m,iw;
    char move[20];
    unsigned char outbuffer[1024];
    unsigned char *s,*o,c,oc,*ob,*sc;
    unsigned tw = width, th = height;

#ifndef NORESIZE
    if (need_resize)
    {
        get_dimz(&tw, &th);
        if (tw > width)  tw = width;
        if (th > height) th = height;
    }
#endif

    s=screen; o=oscreen;
    
    *move=0;

    for (i=iw=0; i<th; i++,iw+=width) 
    {
        ob=outbuffer;
        for (j=0; j<tw; j++,s++,o++) 
        {
            if (*s<' ' /*|| *s>126*/) { *s+=' '; }
            //*s += (*s < ' ') << 5;
            if (*s!=*o) {
                *o=*s;
                l=sgo(move,cx,cy,i+1,j+1);
                if ((cx!=(i+1)) ||
                   (((cx==(i+1)) && (((j-cy)>l) || (cy>=(j+1)))))) {
                    /*printf("%s",move);*/
                    for (m=0; m<l; m++)
                        *ob++=move[m];

                    cx=i+1; cy=j+1;
                    bytes+=l; 
                } else {
                    sc = screen+iw+cy-1;
                    memcpy(ob,sc,j-cy+1);
                    ob += j-cy+1;
                    bytes += j-cy+1;
                    cy = j+1;
                }
                *ob++=*s;
                cy++; bytes++;
            }
        }
        s += width - tw;
        o += width - tw;
        fwrite(outbuffer,1,ob-outbuffer,stdout);
    }

    if (showstats) {
        go(th+1,1); 
        fflush(stdout);
        ++frame;
        fprintf(stderr,"Frame: %8u  Bytes: %8u  Bytes/Frame: %8u\r",frame, bytes, bytes/frame);
    } else {
        fflush(stdout);
        ++frame;
    }

#ifndef NORESIZE
    if (can_resize && need_resize)
        do_resize();
#endif
}

void srefresh(void)
{
    int* s;
    int i;

    s=(int *)oscreen;

    for (i=whsize>>2;i>0;i--) *s++=0x20202020;

    bytes+=printf("\033[H\033[2J");
    supdate();  
}

void sdump(void)
{
    int i,j,k,l;

    printf("\033c\033[H\033[2J");
    for (i=0; i<height; i++) {
        for (j=0; j<width; j++) {
            k=j+i*width;
            putchar(screen[k]);
        }
        putchar('\n');
    }
    fflush(stdout);
    while(getchar()!='\n') ;
    printf("\033c\033[H\033[2J");
    for (i=0; i<height; i++) {
        for (j=0; j<width; j++) {
            k=j+i*width;
            putchar(oscreen[k]);
        }
        putchar('\n');
    }
    fflush(stdout);

}

box (char c, int ulx, int uly, int lrx, int lry)
{
    int i,j,k;
    for (i=ulx-1; i<lrx; i++)
        for (j=uly-1; j<lry; j++)
            screen[i*width+j]=c;
}
    
clearbox(char c,int ulx, int uly, int lrx, int lry)
{
    int i,j,k,l;
    for (l=width>>1; l>0; l--) {
        for (i=ulx-1; i<lrx; i++) {
            for (j=uly-1; j<lry; j++) {
                k=i*width+j;
                screen[k]=((((i+width)%l)==0)&&(((j+width)%l)==0))?c:screen[k];
            }
        }
        supdate();
    }
}

void vline(char c, int col, int s, int e)
{
    unsigned char *ss;
    if (s>e) {s^=e^=s^=e;}
    ss=screen+col+s*width-1;
    e-=s-1;
    while(e--) {*ss=c;ss+=width;}
}

void hline(char c, int row, int s, int e)
{
    unsigned char *ss;
    if (s>e) {s^=e^=s^=e;}
    ss=screen+s+row*width-1;
    e-=s-1;
    while(e--) {*ss++=c;}
}


#define absv(a) (((a)<0)?-(a):(a))

#if 0
void line(char c,float ulx,float uly,float lrx,float lry)
{
    int k,l;
    screen[(l=uly-1)+(k=ulx-1)*width]=c;
    screen[(l=lry-1)+(k=lrx-1)*width]=c;
    if ((absv(lrx-ulx))<=1 && (absv(lry-uly)<=1)) {
        return;
    }
    /*printf("%c %d %d %d %d\n",c,ulx,uly,lrx,lry);*/
    line(c, ulx, uly, (ulx+lrx)/2.0, (uly+lry)/2.0);
    line(c, (ulx+lrx)/2.0, (uly+lry)/2.0, lrx, lry);
}
#endif


/*
** Public Domain mode 13h Bresenham line/circle algorithms
** By Brian Dessent
**
** Written for Borland, modified for others by Bob Stout
*/

/* MODIFIED FOR VT.C by Joseph Zbiciak */

void bresenham_line(int x, int y, int x2, int y2, char c);
void bresenham_circle(int xc, int yc, int r, char c);


#define line(a,b,c,d,e)  bresenham_line((c)-1,(b)-1,(e)-1,(d)-1,(a))
#define circle(a,b,c,d)  bresenham_circle((b),(c),(d),(a))

/* code begins */

/* plots a dot at (x, y) with color c */

void putpixel(int x, int y, char c) 
{
    if (x<0 || y<0 || x>=width || y>=height || c==XPARENT) { return; }
    screen[x+y*width]=c;
    return;
}

#define putpixel_(x,y,c) if (((c)==XPARENT||(x)<0||(y)<0||(x)>=width||(y)>=height)) ;else screen[(x)+(y)*width]=(c)
#define putpixel(x,y,c) if ((c)!=XPARENT) screen[(x)+(y)*width]=(c)
#define putpixel2(x,y,c) screen[(x)+(y)*width]=(c)
#define plotdot(x,y,c) putpixel_(x,y,c)



/* draws a line from (x, y) to (x2, y2) in color c */

void bresenham_line(int x, int y, int x2, int y2, char c)
{
      int i, steep = 0, sx, sy, dx, dy, e;

      dx = abs(x2 - x);
      sx = ((x2 - x) > 0) ? 1 : -1;
      dy = abs(y2 - y);
      sy = ((y2 - y) > 0) ? 1 : -1;

      if(dy > dx)
      {
        steep = 1;
        x ^= y;  /* swap x and y */
        y ^= x;
        x ^= y;
        dx ^= dy; /* swap dx and dy */
        dy ^= dx;
        dx ^= dy;
        sx ^= sy; /* swap sx and sy */
        sy ^= sx;
        sx ^= sy;
      }

      e = 2 * dy - dx;
      if (steep) {
        for(i = 0;i < dx;i++)
        {
            plotdot(y, x, c);
            while(e >= 0)
            {
                y += sy;
                e -= 2 * dx;
            }
            x += sx;
            e += 2 * dy;
        }
      } else {
        for(i = 0;i < dx;i++)
        {
            plotdot(x, y, c);
            while(e >= 0)
            {
                y += sy;
                e -= 2 * dx;
            }
            x += sx;
            e += 2 * dy;
        }
      }
      plotdot(x2, y2, c);
}

/* draws a circle at (xc, yc) with radius r in color c
**
** note: the scaling factor of (SCREEN_WIDTH / SCREEN_HEIGTH) is used when
** updating d.  This makes round circles.  If you want ellipses, you can
** modify that ratio.
*/

void bresenham_circle(int xc, int yc, int r, char c)
{
      int x = 0, y = r, d = 2 * (1 - r);

      while(y > 0)
      {
        plotdot(xc + x, yc + y, c);
        plotdot(xc + x, yc - y, c);
        plotdot(xc - x, yc + y, c);
        plotdot(xc - x, yc - y, c);
        if(d + y > 0)
        {
          y -= 1;
          /*d -= (2 * y * SCREEN_WIDTH / SCREEN_HEIGTH) - 1;*/
          d -= (2*y*width/height)-1;
        }
        if(x > d)
        {
          x += 1;
          d += (2 * x) + 1;
        }
      }
}

void setmode(int i)
{
}

typedef struct /* winsize */ {
    unsigned short  ws_row;         /* rows, in characters */
    unsigned short  ws_col;         /* columns, in characters */
    unsigned short  ws_xpixel;      /* horizontal size - not used */                
    unsigned short  ws_ypixel;      /* vertical size - not used */
} Dims;

void initscreen(int w, int h, int st) 
{
    int i,j,k;
    unsigned int *s,*o;

#ifndef NORESIZE
    if (h==-1 || w==-1)
    {
        Dims mywinz;
        /* use autodimension setting */
        if ( ioctl(0, TIOCGWINSZ, &mywinz) == 0 )
        {
            void vt_sig_winch(int);

            h=mywinz.ws_row;
            w=mywinz.ws_col;
            signal(SIGWINCH,vt_sig_winch);
        } else 
        {
            h=w=0;
            errno=0;
        }

        if (h<2 || w<2)
        {
            fprintf(stderr,"Unable to get window size.\n");
            if (errno==0)
                fprintf(stderr,"Try using 'stty rows [height] cols [width]'\n");
            else
                perror("newvt: initscreen()");
            exit(1);
        }
    }
#else
    if (h==-1 || w==-1)
    {
        fprintf(stderr,"Autosizing disabled in this compilation!\n");
        exit(1);
    }
#endif


    height=--h; width=--w;
    whsize=width*height+4;
    showstats=st;

    s=(unsigned int*) (screen=(unsigned char *)malloc(sizeof(char)*whsize));
    o=(unsigned int*) (oscreen=(unsigned char *)malloc(sizeof(char)*whsize));
    i=whsize/sizeof(unsigned int);
    while (i--) { *s++=*o++=0x20202020; }

    fputs(TITLE", "VERSION", by "AUTHOR,stdout);
    sleep(1);
    fputs("\033[H\033[2J",stdout);
    go(1,1);
    supdate();

    return;
}

void closescreen(void)
{
    go(height,1);
    free(screen);
    free(oscreen);
    return;
}

#ifndef NORESIZE
void do_resize(void)
{
    char * s, *o;
    Dims mywinz;
    int i,j;

    need_resize = 0;

    if (ioctl(0,TIOCGWINSZ,&mywinz)!=0)
    {
        fprintf(stderr,"Unable to get window size.\n");
        perror("newvt: vt_sig_winch");
        exit(1);
    }

    mywinz.ws_row--;
    mywinz.ws_col--;

/*
    go(mywinz.ws_row - 1, 1);
    fflush(stdout);
    fprintf(stderr,"Resize [%4d,%4d] -> [%4d,%4d]",
            width,height,mywinz.ws_col,mywinz.ws_row);
    go(1,1);
*/

    whsize=(mywinz.ws_row)*(mywinz.ws_col)+4;

    s=(char *)malloc(sizeof(char)*whsize);
    o=(char *)malloc(sizeof(char)*whsize);

    for (j=0;j<height && j<mywinz.ws_row;j++)
        for (i=0;i<width && i<mywinz.ws_col;i++) 
        {
            s[j*mywinz.ws_col+i]= screen[j*width+i];
            o[j*mywinz.ws_col+i]= screen[j*width+i]-1;
        }

    if (mywinz.ws_col>width)
    {
        for (j=0;j<mywinz.ws_row;j++)
            for (i=width;i<mywinz.ws_col;i++)
            {
                s[j*mywinz.ws_col+i]=' ';
                o[j*mywinz.ws_col+i]='!';
            }
    }

    if (mywinz.ws_row>height)
    {
        for (j=height;j<mywinz.ws_row;j++)
            for (i=0; i<mywinz.ws_col; i++)
            {
                s[j*mywinz.ws_col+i]=' ';
                o[j*mywinz.ws_col+i]='!';
            }
    }

    free(screen);
    free(oscreen);

    screen=s;
    oscreen=o;
    width=mywinz.ws_col;
    height=mywinz.ws_row;


    return;
}

void get_dimz(int *w, int *h)
{
    char * s, *o;
    Dims mywinz;

    if (ioctl(0,TIOCGWINSZ,&mywinz)!=0)
    {
        fprintf(stderr,"Unable to get window size.\n");
        perror("newvt: vt_sig_winch");
        exit(1);
    }

    *w=mywinz.ws_col - 1;
    *h=mywinz.ws_row - 1;
}

void vt_sig_winch(int sig)
{
	static volatile int in_vt_sig_winch = 0;
    int old_can = can_resize;
    signal(SIGWINCH,vt_sig_winch);

    if (!in_vt_sig_winch)
	{
		in_vt_sig_winch = 1;
        need_resize     = 1;
        can_resize      = 0;
        srefresh();
        can_resize      = old_can;
		in_vt_sig_winch = 0;
	}
}
#endif

/* ======================================================================== */
/*  Please see the file "COPYING" for the complete GNU GPL License.         */
/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       */
/*  General Public License for more details.                                */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License       */
/*  along with this program; if not, write to the Free Software             */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               */
/* ======================================================================== */
/*                 Copyright (c) 1995-2000, Joseph Zbiciak                  */
/* ======================================================================== */
