/* Pac-Man for Terminals */

#ifndef _PAC_TYPES_H_
#define _PAC_TYPES_H_

#include "pacman.h"

enum { Init=1, Active, Chased, Dead };
enum { Stopped, Up=1, Down=2, Left=3, Right=4 };

typedef struct tMazeLoc
{   
    int x,y;
} TMazeLoc;

typedef struct tLevel
{
    unsigned char map[MAX_X*MAX_Y];     /* actual map array     */
    int         mazex,mazey;            /* maze dimensions      */
    int         p_speed;                /* speed in moves/10sec */
    int         pc_speed;               /* speed when chomping  */
    TMazeLoc    p_home;                 /* pacman's home        */
    int         numghosts;              /* number of ghosts     */
    int         g_speed [MAXGHOSTS];    /* ghost speed in m/10s */
    TMazeLoc    g_home  [MAXGHOSTS];    /* ghost homes          */
    TMazeLoc    g_init  [MAXGHOSTS];    /* ghost initial goals  */
    TMazeLoc    power_ups[MAXPOWER];    /* powerup locations    */
    int         power_up;               /* Duration of a pwrup  */
    int         tot_power,power_left;   /* # of pwrups in maze  */
    int         tot_dots,dots_left;     /* # of dots in maze    */
    int         sx,sy,dw,dh,lc,tr;   /* display state variables */
    int         redraw;
} TLevel;

typedef struct tMoveInfo
{
    int         l_move;     /* last move (in ticks)     */
    int         speed;      /* speed (int ticks/move)   */
    int         cspeed;     /* speed (int ticks/move)   */
} TMoveInfo;

typedef struct tGhost
{
    int         state;  /* Init, Active, Chased, Dead */
    int         inhome; /* Whether the ghost is "in home" */
    TMazeLoc    loc;    /* Ghost position           */
    TMazeLoc    goal;   /* Ghost goal       */
    TMazeLoc    mem[MEMORY];
    int         mp;
    int         dir;
    TMoveInfo   mv;
} TGhost;

typedef struct tPacman
{
    int         rand;
    int         god;
    int         power,fpower;
    int         g_score;
    TMazeLoc    loc;
    int         dir;
    int         ndir;
    int         score;
    int         lives;
    int         time_left;
    TMoveInfo   mv;
} TPacman;

typedef struct tPlayer
{
    TLevel      lev;
    TPacman     pac;
    TGhost      ghost[MAXGHOSTS];
    int         lv;
} TPlayer;

#ifdef _PACMAN_C_

TPlayer pl      [NUMPLAYERS];
TLevel  level   [NUMLEVELS];

#else

extern TPlayer pl       [NUMPLAYERS];
extern TLevel  level    [NUMLEVELS];

#endif

/* Code simplification macros */
#define P  pl[up]
#define PG pl[up].ghost[g]
#define PP pl[up].pac
#define PL pl[up].lev

/* inlines that work on the above types (mostly) */


static TMazeLoc __pt0, __pt1, __pt2, __pt3, __pt4;

#define point(_x,_y)                            \
    (                                           \
        (__pt0.x=(_x)),                         \
        (__pt0.y=(_y)),                         \
        __pt0                                   \
    )

#define pdir(_a,_d)                             \
    (                                           \
        (__pt1.x=(_a).x + dx[(_d)]),            \
        (__pt1.y=(_a).y + dy[(_d)]),            \
        __pt1                                   \
    )

#define pdiff(_a,_b)                            \
    (                                           \
        (__pt2=(_a)),(__pt3=(_b)),              \
        (__pt4.x=__pt2.x - __pt3.x),            \
        (__pt4.y=__pt2.y - __pt3.y),            \
        __pt4                                   \
    )

#define padd(_a,_b)                             \
    (                                           \
        (__pt2=(_a)),(__pt3=(_b)),              \
        (__pt4.x=__pt2.x + __pt3.x),            \
        (__pt4.y=__pt2.y + __pt3.y),            \
        __pt4                                   \
    )

#define idx(pt)  ((pt).x + (pt).y * PL.mazex)


static inline int pass(TMazeLoc pt, int dir, int state)
{
    const char * ps;
    pt=pdir(pt,dir);
    if (pt.x>=PL.mazex-1) pt.x=0;  /* for tunnels */
    if (pt.x<0)           pt.x=PL.mazex-2;
    ps=passible+(state==Active || state==Chased);
    return 
        strchr(ps,PL.map[idx(pt)  ])!=NULL &&
        strchr(ps,PL.map[idx(pt)+1])!=NULL;
}

static inline const int coinc(const TMazeLoc pt1, const TMazeLoc pt2)
{
    return (pt1.y==pt2.y) && ((pt1.x-pt2.x)<2 && -2<(pt1.x-pt2.x));
}

static inline const int sgn(const int n) { return n>0?1:n<0?-1:0; }


#endif /*_PAC_TYPES_H_*/
/* ======================================================================== */
/*  Please see the file "COPYING" for the complete GNU GPL License.         */
/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       */
/*  General Public License for more details.                                */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License       */
/*  along with this program; if not, write to the Free Software             */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               */
/* ======================================================================== */
/*                 Copyright (c) 1995-2000, Joseph Zbiciak                  */
/* ======================================================================== */
