/* Pac-Man for Terminals */

#define _PACMAN_C_

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include "newvt.h"
#include "pac-types.h"
#include "pacman.h"
#include "clock.h"
#include "keybd.h"
#include "actors.h"
#include "level.h"
#include "intro.h"
#include "digits.h"


void show_logo(void)
{
    static ofs=0;
    int i,j,k,l,r;

    if (!ofs) ofs=Clock();

    r=(height-4)/2;
    j=(Clock()-ofs)/(CLOCKS_PER_SEC/3);

    l=strlen(logo[0]);

    for(k=0;k<4;k++)
    {
        for (i=0;i<(RM-1);i++)
        {
            screen[width-RM+(r+k)*width+i]=logo[k][(j+i)%l];
        }
    }
}

void show_map(int force)
{
    int l,r,t,b;
    int sx,dw,sy,dh;
    int i,j,tr,lc;
    unsigned char *s1,*s2,*s3,*s4;

    t=TM; b=height-BM;
    l=LM; r=width-RM;

    if (b-t>=PL.mazey) 
    { 
        tr=t + ((b-t)-PL.mazey)/2; 
        sy=0; dh=PL.mazey;
    } else
    {
        tr=t;
        sy=PP.loc.y-(b-t)/2;
        if (sy<0) sy=0;
        if (sy>(PL.mazey-(b-t))) sy=PL.mazey-(b-t);
        dh=b-t;
    }

    if (r-l>=PL.mazex) 
    { 
        lc=l + ((r-l)-PL.mazex)/2; 
        sx=0; dw=PL.mazex;
    } else
    {
        lc=l;
        sx=PP.loc.x-(r-l)/2;
        if (sx<0) sx=0;
        if (sx>(PL.mazex-(r-l))) sx=PL.mazex-(r-l);
        dw=r-l;
    }

    for (i=0;i<PL.tot_power;i++)
    {
        int x,y;
        x=PL.power_ups[i].x-sx; y=PL.power_ups[i].y-sy;
        if (x<0 || y<0 || x>dw || y>dh) continue;
        if ((Clock()/(CLOCKS_PER_SEC/2))&1)
        {
            putpixel(x+lc,y+tr,' ');
        } else
        {
            putpixel(x+lc,y+tr,PL.map[idx(PL.power_ups[i])]);
        }
    }

    if (!force && PL.sx>=0 && abs(PL.sx-sx)<1 && abs(PL.sy-sy)<1) return;

    PL.sx=sx; PL.sy=sy;
    PL.dw=dw; PL.dh=dh;
    PL.lc=lc; PL.tr=tr;
    PL.redraw=Clock();

    s1=screen+lc+tr*width;
    s2=PL.map+idx(point(sx,sy));


    for(i=0;i<dh;i++)
    {
        s3=s1;
        s4=s2;
        for(j=0;j<dw;j++) *s1++=*s2++;
        s1=s3+width;
        s2=s4+PL.mazex;
    }

}

void draw_mnum(int num,int sx, int sy, int digs)
{
    /* draw medium-sized number at screen coords (sx,sy) */
    int i,j,k,l,nz=0;

    l=sx;
    for (i=digs-1;i>=0;i--)
    {
        j=1; k=i;
        while (k-->0) j*=10;    
    
        j=(num/j)%10;
        if (j!=0 || nz)
        {
            for (k=0;k<3;k++)
            {
                screen[l+0+(k+sy)*width]=digits_med[j][k][0];
                screen[l+1+(k+sy)*width]=digits_med[j][k][1];
                screen[l+2+(k+sy)*width]=digits_med[j][k][2];
            }
            nz=1;
        } else
        {
            for (k=0;k<3;k++) for(j=0;j<3;j++) screen[l+j+(k+sy)*width]=' ';
        }

        l+=3;
    }
}

void show_stats(p)
{
    static unsigned int user_time=0, real_time=0, sys_time=0;
    static unsigned int o_frames=0,o_bytes=0,fr_time=0,sfr_time=0;
    static double pk_cps=0.0;
    unsigned int nu,ns,nr;
    struct tms t;
    int i,j;
    int tp;
    char buf[100];

    tp=height/2-10;
#if 0
    strcpy(buf,"Score:");
    for (i=0;i<6;i++) screen[(tp)*width+2+i]=buf[i];
    sprintf(buf,"%8d",PP.score);
    for (i=0;i<8;i++) screen[(tp+1)*width+2+i]=buf[i];
#endif
    if (PP.rand && hi_score)
        draw_mnum(hi_score,(width-8*3)/2,0,8);
    else
        draw_mnum(PP.score,(width-8*3)/2,0,8);

    for (j=0;j<4;j++)   
    for (i=0;i<4;i++)
        if (PP.lives>(i+j*4+1))
        {
            putpixel(i*3+2,j*2+tp+6,'(');
            putpixel(i*3+3,j*2+tp+6,'<');
        } else
        {
            putpixel(i*3+2,j*2+tp+6,' ');
            putpixel(i*3+3,j*2+tp+6,' ');
        }

    if (PP.power>0)
    {
        strcpy(buf,"Pac Power:");
        for (i=0;i<10;i++) screen[(tp+16)*width+2+i]=buf[i];
        j=(10*PP.power)/PP.fpower;
        for (i=0;i<10;i++) screen[(tp+17)*width+2+i]=(i<=j)?'#':'|';
    } else if (PP.fpower!=0)
    {
        for (i=0;i<10;i++) screen[(tp+16)*width+2+i]=' ';
        for (i=0;i<10;i++) screen[(tp+17)*width+2+i]=' ';
        PP.fpower=0;
    }

    sprintf(buf,"sync: %7s",sync_type==NO_SYNC?"none":sync_type==BG_SYNC?"bg":sync_type==HARD_SYNC?"hard":sync_type==SOFT_SYNC?"soft":"kitchen");


    for (i=0;i<13;i++) screen[3*width-15+i]=buf[i];
    /*show_logo();*/

    times(&t);
    nu=t.tms_utime;
    ns=t.tms_stime;
    nr=Clock_unad();

    if (nr > real_time)
    {
        sprintf(buf,"CPU User: %3d%%",(100*(nu-user_time))/(nr-real_time));
        for (i=0;i<14;i++) screen[5*width-15+i]=buf[i];

        sprintf(buf,"CPU Sys:  %3d%%",(100*(ns- sys_time))/(nr-real_time));
        for (i=0;i<14;i++) screen[6*width-15+i]=buf[i];
    }
#ifdef CPS
    if ((frame-o_frames)>=16)
    {
        double cps;
        sprintf(buf,"FPS:     %5.2f",((double)(frame-o_frames)*CLOCKS_PER_SEC)/((double)(nr-fr_time)));
        for (i=0;i<14;i++) screen[7*width-15+i]=buf[i];
        sprintf(buf,"CPS:   %7.2f",cps=(((double)(bytes-o_bytes)*(double)CLOCKS_PER_SEC)/((double)(nr-fr_time))));
        for (i=0;i<14;i++) screen[8*width-15+i]=buf[i];
        if (cps>pk_cps) pk_cps=cps;
        sprintf(buf,"Pk CPS:%7.2f",pk_cps);
        for (i=0;i<14;i++) screen[9*width-15+i]=buf[i];
        sprintf(buf,"Av CPS:%7.2f",((double)(bytes)*CLOCKS_PER_SEC)/((double)(nr-sfr_time)));
        for (i=0;i<14;i++) screen[10*width-15+i]=buf[i];
        o_frames=frame;
        o_bytes =bytes;
        if (!sfr_time) sfr_time=nr;
        fr_time=nr;
    }
#endif

    if (Clock()&7==0 || real_time==0) 
    {
        real_time=nr;
        user_time=nu;
        sys_time =ns;
    }
    
}


int game_loop(void)
{
    int died=0,i,j,x,y,g,pu=0,do_sync=0,fl=0,update=0,fc_sync=0;
    can_resize = 0;

    if (width < 40 || height < 20)
    {
        fprintf(stderr,"\033[H\033[2JWindow too small.\n");
        return -1;  /* abort */
    }

    PL.sx=PL.sy=-1;
    show_map(1);

    while (!died && PL.dots_left && PP.time_left>0)
    {
        if (fc_sync<=0 && sync_type==BG_SYNC) 
        {       
            printf("\033[6n"); fflush(stdout);
            fc_sync=10;
        } else if (sync_type==BG_SYNC) fc_sync--;
        usleep(20000);
        if ((i=parse_key())==-1) return -1;
        if (i==-2) return -2;

        hide_all();

        if (i<=-7) { P.lv=-8-i; PL.dots_left=0; return 0; }

        for (g=0;g<PL.numghosts;g++)
        {
            if (Clock()>PG.mv.l_move+PG.mv.speed)
            {
                int mv;
                move_ghost(g);
                update=1;
                mv=PG.mv.speed;
                if (PG.state==Dead) mv=(mv*2)/3;
                PG.mv.l_move+=mv;
                if (is_tunnel(PG.loc))
                {
                    PG.mv.l_move+=mv;
                }
                if (Clock()>PG.mv.l_move+PG.mv.speed) PG.mv.l_move=Clock();
            }
            /*show_ghost(g);*/
        }


        for (g=0;g<PL.numghosts;g++)
        {
            if (PG.dir==PP.dir) continue;
            if (coinc(PG.loc,PP.loc))
            {
                if (PG.state==Chased)
                {
                    PG.state=Dead;
                    PG.goal =PL.g_home[g];
                    PP.score+=PP.g_score;
                    PP.g_score*=2;
                } else if (PG.state!=Dead)
                {
                    PP.lives-=(died=!PP.god);
                }
            }
        }
    
        if (!died && Clock()>=PP.mv.l_move+PP.mv.speed)
        {
            do_sync++;
            move_pacman();
            update=1;
            show_map(0);
            if (PP.rand) PP.time_left--;
            PP.mv.l_move+=PP.mv.speed;
            if (fl)
            {
                PP.mv.l_move+=PP.mv.cspeed-PP.mv.speed;
            }
            if (Clock()>PP.mv.l_move+PP.mv.speed) PP.mv.l_move=Clock();
    
            for(j=0;j<2;j++)
            {
                TMazeLoc l;

                l=padd(PP.loc,point(j,0));
    
                switch (PL.map[idx(l)])
                {
                    case DOT:
                        PL.map[idx(l)]=SPACE;
                        PP.score+=(DOT_SCORE*(P.lv+2))/2;
                        PL.dots_left--;
                        if (!fl) PP.mv.l_move+=PP.mv.cspeed-PP.mv.speed;
                        fl=2;
#ifdef NOISY
                        fputc('\007',stderr); fflush(stderr);
#endif
                        break;
                    case POWER:
                        PL.map[idx(l)]=SPACE;
                        PP.score+=(POWER_SCORE*(P.lv+2))/2;
                        PP.power=(PL.power_up*(2*PL.tot_power-pu))/(2*PL.tot_power);
                        PP.fpower=PP.power;
                        PL.power_left--;
                        PP.g_score=(GHOST_SCORE*(P.lv+2))/2;
                        for (g=0;g<PL.numghosts;g++)
                        {
                            int dir;
                            if (PG.state==Dead) continue;
    
                            dir=PG.dir;
                            PG.dir=dir==Up?Down:
                                        (dir==Down?Up:
                                        (dir==Left?Right:
                                        Left));
                            PG.state=Chased;
                        }
                        break;
                    default:
                        break;
                }
            }
            if (fl) fl--;   
    
            for (g=0;g<PL.numghosts;g++)
            {
                if (coinc(PG.loc,PP.loc))
                {
                    if (PG.state==Chased)
                    {
                        PG.state=Dead;
                        PG.goal =PL.g_home[g];
                        PP.score+=PP.g_score;
                        PP.g_score*=2;
                    } else if (PG.state!=Dead)
                    {
                        PP.lives-=(died=!PP.god);
                    }
                }
            }

            if (PP.power && --PP.power==0)
                for (g=0;g<PL.numghosts;g++) if (PG.state==Chased && !PG.inhome)
                {
                    PG.state=Active;
                    PG.goal =PP.loc;
                }
        }

        for (g=0;g<PL.numghosts;g++) show_ghost(g);
        show_pacman();

        show_stats();
        if (update || (sync_type==BG_SYNC && rcv_sync))
        {
            if (sync_type==BG_SYNC && rcv_sync)
            {
                oscreen[width*height-1]='!';
                screen [width*height-1]=' ';
                supdate();
                rcv_sync=0;
                printf("\033[6n"); fflush(stdout);
            } else if (sync_type!=BG_SYNC)
            {
                oscreen[width*height-1]='!';
                screen [width*height-1]=' ';
                supdate();
                fflush(stdout);
            }
            if ((sync_type==HARD_SYNC && do_sync>=1) ||
                (sync_type==SOFT_SYNC && do_sync>=2) || 
                 sync_type==KITCHEN_SYNC)
            {
                sync_via_rcp();
                do_sync=0;
            }
                         
            if (need_resize)
            {
                int i;

                for (i=0; i<width * height; i++)
                {
                    screen[i]  = 0x20;
                    oscreen[i] = '!';
                }

                can_resize = 1;
                srefresh();
                if (width < 40 || height < 20)
                {
                    fprintf(stderr,"\033[H\033[2JWindow too small.\n");
                    return -1;  /* abort */
                }
                can_resize = 0;
                show_map(1);
                hide_all();
                for (g = 0; g < PL.numghosts; g++) show_ghost(g);
                show_pacman();
                show_stats();
                supdate();
            }
        }

        if (!PP.rand && PP.score>hi_score) hi_score=PP.score;
    }

    if (!PP.rand && PP.score>hi_score) hi_score=PP.score;
    sleep(1);
    return died;
}

int main()
{
    int i,gl;

    srand(time(0));
    initscreen(-1,-1,0);
    initkbd();
    can_resize = 0;
    
    load_levels();

    do
    {
        init_pacman();
        init_ghosts();
    
        if (title_screen() == -1) break;

        up=0;
        PP.score=0;
    
        P.lv=0;
        while (PP.lives>0)
        {
            init_level(P.lv);
            while (PL.dots_left && PP.lives>0)
            {
                reset_pacman();
                reset_ghosts();
                open_level(P.lv);
                flushkbd();
                if ((gl=game_loop())>0)
                    kill_pacman();
                if (gl<0) break;
            }
            if (gl<0) break;
            if (!PL.dots_left) 
            {
                close_level(P.lv);
                P.lv++;
                if (P.lv>=NUMLEVELS) P.lv=0;
            }
        }
        if (gl>=0) game_over(P.lv);
    } while (gl>=0);

    restorekbd();
    go(height-1,1);
    printf("Thank you for playing VT-100 Pac-Man!\n");
    closescreen();
    return 0;
}   
/* ======================================================================== */
/*  Please see the file "COPYING" for the complete GNU GPL License.         */
/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       */
/*  General Public License for more details.                                */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License       */
/*  along with this program; if not, write to the Free Software             */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               */
/* ======================================================================== */
/*                 Copyright (c) 1995-2000, Joseph Zbiciak                  */
/* ======================================================================== */
